(function () {

/* Imports */
var Meteor = Package.meteor.Meteor;
var MongoInternals = Package.mongo.MongoInternals;
var Mongo = Package.mongo.Mongo;
var Random = Package.random.Random;
var check = Package.check.check;
var Match = Package.check.Match;
var _ = Package.underscore._;
var Accounts = Package['accounts-base'].Accounts;
var HTTP = Package.http.HTTP;
var HTTPInternals = Package.http.HTTPInternals;

/* Package-scope variables */
var SandstormDb, Packages, DevApps, UserActions, Grains, RoleAssignments, Contacts, Sessions, SignupKeys, ActivityStats, DeleteStats, FileTokens, ApiTokens, Notifications, StatsTokens, Misc, Settings, Migrations, StaticAssets, AssetUploadTokens, Plans, isDemoUser, isSignedUp, isSignedUpOrDemo, isUserOverQuota, isUserExcessivelyOverQuota, isAdmin, isAdminById, findAdminUserForToken, matchWildcardHost, makeWildcardHost, getWildcardOrigin, allowDevAccounts, roleAssignmentPattern, addStaticAsset, userPictureUrl, fetchPicture;

(function () {

////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                    //
// packages/sandstorm-db/db.js                                                                        //
//                                                                                                    //
////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                      //
// Sandstorm - Personal Cloud Sandbox                                                                 // 1
// Copyright (c) 2014 Sandstorm Development Group, Inc. and contributors                              // 2
// All rights reserved.                                                                               // 3
//                                                                                                    // 4
// Licensed under the Apache License, Version 2.0 (the "License");                                    // 5
// you may not use this file except in compliance with the License.                                   // 6
// You may obtain a copy of the License at                                                            // 7
//                                                                                                    // 8
//   http://www.apache.org/licenses/LICENSE-2.0                                                       // 9
//                                                                                                    // 10
// Unless required by applicable law or agreed to in writing, software                                // 11
// distributed under the License is distributed on an "AS IS" BASIS,                                  // 12
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                           // 13
// See the License for the specific language governing permissions and                                // 14
// limitations under the License.                                                                     // 15
                                                                                                      // 16
// This file defines the database schema.                                                             // 17
                                                                                                      // 18
// Useful for debugging: Set the env variable LOG_MONGO_QUERIES to have the server write every        // 19
// query it makes, so you can see if it's doing queries too often, etc.                               // 20
if (Meteor.isServer && process.env.LOG_MONGO_QUERIES) {                                               // 21
  var oldFind = Mongo.Collection.prototype.find;                                                      // 22
  Mongo.Collection.prototype.find = function () {                                                     // 23
    console.log(this._prefix, arguments);                                                             // 24
    return oldFind.apply(this, arguments);                                                            // 25
  }                                                                                                   // 26
}                                                                                                     // 27
                                                                                                      // 28
// Users = new Mongo.Collection("users");                                                             // 29
// The users collection is special and can be accessed through `Meteor.users`.                        // 30
// See https://docs.meteor.com/#/full/meteor_users. Entries in the users collection correspond to     // 31
// accounts. Each represents a capability store belonging to a human.                                 // 32
//                                                                                                    // 33
// Note that accounts are not the same thing as identities. An account may have multiple identities   // 34
// attached to it. Identities provide a way to authenticate the user and are also used to present     // 35
// globally unique and stable user indentifiers to grains via `Grain.UserInfo.userId`.                // 36
//                                                                                                    // 37
// Each entry in this collection contains:                                                            // 38
//   _id: Random.                                                                                     // 39
//   createdAt: Date when this entry was added to the collection.                                     // 40
//   lastActive: Date of the user's most recent interaction with this Sandstorm server.               // 41
//   profile: Object containing profile data editable by the user and visible to other users. May     // 42
//            include the following fields. Note that if any field is missing, the first fallback     // 43
//            is to check `services` for details provided by the identity provider (the details       // 44
//            of which differ per-provider). Only if that is also missing do we fall back to          // 45
//            defaults.                                                                               // 46
//       name: String containing the display name of the user. Default: first part of email.          // 47
//       handle: String containing the user's preferred handle. Default: first part of email.         // 48
//       picture: _id into the StaticAssets table for the user's picture. Default: identicon.         // 49
//       pronoun: One of "male", "female", "neutral", or "robot". Default: neutral.                   // 50
//   services: Object containing login and identity data used by Meteor authentication services.      // 51
//   isAdmin: Boolean indicating whether this user is allowed to access the Sandstorm admin panel.    // 52
//   signupKey: If this is an invited user, then this field contains their signup key.                // 53
//   signupNote: If the user was invited through a link, then this field contains the note that the   // 54
//               inviter admin attached to the key.                                                   // 55
//   signupEmail: If the user was invited by email, then this field contains the email address that   // 56
//                the invite was sent to.                                                             // 57
//   plan: _id of an entry in the Plans table which determines the user's qutoa.                      // 58
//   storageUsage: Number of bytes this user is currently storing.                                    // 59
//   expires: Date when this user's account should be deleted. Only present for demo users.           // 60
//   isAppDemoUser: True if this is a demo user who arrived via an /appdemo/ link.                    // 61
//   payments: Object defined by payments module, if loaded.                                          // 62
//   dailySentMailCount: Number of emails sent by this user today; used to limit spam.                // 63
                                                                                                      // 64
Packages = new Mongo.Collection("packages");                                                          // 65
// Packages which are installed or downloading.                                                       // 66
//                                                                                                    // 67
// Each contains:                                                                                     // 68
//   _id:  128-bit prefix of SHA-256 hash of spk file, hex-encoded.                                   // 69
//   status:  String.  One of "download", "verify", "unpack", "analyze", "ready", "failed", "delete"  // 70
//   progress:  Float.  -1 = N/A, 0-1 = fractional progress (e.g. download percentage),               // 71
//       >1 = download byte count.                                                                    // 72
//   error:  If status is "failed", error message string.                                             // 73
//   manifest:  If status is "ready", the package manifest.  See "Manifest" in package.capnp.         // 74
//   appId:  If status is "ready", the application ID string.  Packages representing different        // 75
//       versions of the same app have the same appId.  The spk tool defines the app ID format        // 76
//       and can cryptographically verify that a package belongs to a particular app ID.              // 77
//   shouldCleanup:  If true, a reference to this package was recently dropped, and the package       // 78
//       collector should at some point check whether there are any other references and, if not,     // 79
//       delete the package.                                                                          // 80
//   url:  When status is "download", the URL from which the SPK can be obtained, if provided.        // 81
                                                                                                      // 82
DevApps = new Mongo.Collection("devapps");                                                            // 83
// List of applications currently made available via the dev tools running on the local machine.      // 84
// This is normally empty; the only time it is non-empty is when a developer is using the spk tool    // 85
// on the local machine to publish an under-development app to this server. That should only ever     // 86
// happen on developers' desktop machines.                                                            // 87
//                                                                                                    // 88
// While a dev app is published, it automatically appears as installed by every user of the server,   // 89
// and it overrides all packages with the same application ID. If any instances of those packages     // 90
// are currently open, they are killed and reset on publish.                                          // 91
//                                                                                                    // 92
// When the dev tool disconnects, the app is automatically unpublished, and any open instances        // 93
// are again killed and refreshed.                                                                    // 94
//                                                                                                    // 95
// Each contains:                                                                                     // 96
//   _id:  The application ID string (as with Packages.appId).                                        // 97
//   packageId:  The directory name where the dev package is mounted.                                 // 98
//   timestamp:  Time when the package was last updated. If this changes while the package is         // 99
//     published, all running instances are reset. This is used e.g. to reset the app each time       // 100
//     changes are made to the source code.                                                           // 101
//   manifest:  The app's manifest, as with Packages.manifest.                                        // 102
                                                                                                      // 103
UserActions = new Mongo.Collection("userActions");                                                    // 104
// List of actions that each user has installed which create new grains.  Each app may install        // 105
// some number of actions (usually, one).                                                             // 106
//                                                                                                    // 107
// Each contains:                                                                                     // 108
//   _id:  random                                                                                     // 109
//   userId:  User who has installed this action.                                                     // 110
//   packageId:  Package used to run this action.                                                     // 111
//   appId:  Same as Packages.findOne(packageId).appId; denormalized for searchability.               // 112
//   appTitle:  Same as Packages.findOne(packageId).manifest.appTitle.defaultText; denormalized so    // 113
//       that clients can access it without subscribing to the Packages collection.                   // 114
//   appVersion:  Same as Packages.findOne(packageId).manifest.appVersion; denormalized for           // 115
//       searchability.                                                                               // 116
//   appMarketingVersion:  Human-readable presentation of the app version, e.g. "2.9.17"              // 117
//   title:  Human-readable title for this action, e.g. "New Spreadsheet".                            // 118
//   command:  Manifest.Command to run this action (see package.capnp).                               // 119
                                                                                                      // 120
Grains = new Mongo.Collection("grains");                                                              // 121
// Grains belonging to users.                                                                         // 122
//                                                                                                    // 123
// Each contains:                                                                                     // 124
//   _id:  random                                                                                     // 125
//   packageId:  _id of the package of which this grain is an instance.                               // 126
//   appId:  Same as Packages.findOne(packageId).appId; denormalized for searchability.               // 127
//   appVersion:  Same as Packages.findOne(packageId).manifest.appVersion; denormalized for           // 128
//       searchability.                                                                               // 129
//   userId:  User who owns this grain.                                                               // 130
//   title:  Human-readable string title, as chosen by the user.                                      // 131
//   lastUsed:  Date when the grain was last used by a user.                                          // 132
//   private: If true, then knowledge of `_id` does not suffice to open this grain.                   // 133
//   cachedViewInfo: The JSON-encoded result of `UiView.getViewInfo()`, cached from the most recent   // 134
//                   time a session to this grain was opened.                                         // 135
//                                                                                                    // 136
// The following fields *might* also exist. These are temporary hacks used to implement e-mail and    // 137
// web publishing functionality without powerbox support; they will be replaced once the powerbox     // 138
// is implemented.                                                                                    // 139
//   publicId:  An id used to publicly identify this grain. Used e.g. to route incoming e-mail and    // 140
//       web publishing. This field is initialized when first requested by the app.                   // 141
                                                                                                      // 142
RoleAssignments = new Mongo.Collection("roleAssignments");                                            // 143
// *OBSOLETE* Before `user` was a variant of ApiTokenOwner, this collection was used to store edges   // 144
// in the permissions sharing graph. This functionality has been subsumed by the ApiTokens            // 145
// collection.                                                                                        // 146
                                                                                                      // 147
Contacts = new Mongo.Collection("contacts");                                                          // 148
// Edges in the social graph.                                                                         // 149
//                                                                                                    // 150
// If Alice has Bob as a contact, then she is allowed to see Bob's profile information and Bob        // 151
// will show up in her user-picker UI for actions like share-by-identity.                             // 152
//                                                                                                    // 153
// Contacts are not symmetric. Bob might be one of Alice's contacts even if Alice is not one of       // 154
// Bob's.                                                                                             // 155
//                                                                                                    // 156
// Each contains:                                                                                     // 157
//   _id: random                                                                                      // 158
//   ownerId: The `_id` of the user who owns this contact.                                            // 159
//   userId:  The `_id` of the contacted user.                                                        // 160
//   petname: Human-readable label chosen by and only visible to the owner. Uniquely identifies       // 161
//            the contact to the owner.                                                               // 162
//   created: Date when this contact was created.                                                     // 163
                                                                                                      // 164
Sessions = new Mongo.Collection("sessions");                                                          // 165
// UI sessions open to particular grains.  A new session is created each time a user opens a grain.   // 166
//                                                                                                    // 167
// Each contains:                                                                                     // 168
//   _id:  random                                                                                     // 169
//   grainId:  _id of the grain to which this session is connected.                                   // 170
//   hostId: ID part of the hostname from which this grain is being served. I.e. this replaces the    // 171
//       '*' in WILDCARD_HOST.                                                                        // 172
//   timestamp:  Time of last keep-alive message to this session.  Sessions time out after some       // 173
//       period.                                                                                      // 174
//   userId:  User who owns this session.                                                             // 175
//   hashedToken: If the session is owned by an anonymous user, the _id of the entry in ApiTokens     // 176
//       that was used to open it. Note that for old-style sharing (i.e. when !grain.private),        // 177
//       anonymous users can get access without an API token and so neither userId nor hashedToken    // 178
//       are present.                                                                                 // 179
//   powerboxView: If present, this is a view that should be presented as part of a powerbox          // 180
//       interaction.                                                                                 // 181
//     offer: The webkey that corresponds to cap that was passed to the `offer` RPC.                  // 182
                                                                                                      // 183
SignupKeys = new Mongo.Collection("signupKeys");                                                      // 184
// Invite keys which may be used by users to get access to Sandstorm.                                 // 185
//                                                                                                    // 186
// Each contains:                                                                                     // 187
//   _id:  random                                                                                     // 188
//   used:  Boolean indicating whether this key has already been consumed.                            // 189
//   note:  Text note assigned when creating key, to keep track of e.g. whom the key was for.         // 190
//   email: If this key was sent as an email invite, the email address to which it was sent.          // 191
                                                                                                      // 192
ActivityStats = new Mongo.Collection("activityStats");                                                // 193
// Contains usage statistics taken on a regular interval. Each entry is a data point.                 // 194
//                                                                                                    // 195
// Each contains:                                                                                     // 196
//   timestamp: Date when measurements were taken.                                                    // 197
//   daily: Contains stats counts pertaining to the last day before the sample time.                  // 198
//   weekly: Contains stats counts pertaining to the last seven days before the sample time.          // 199
//   monthly: Contains stats counts pertaining to the last thirty days before the timestamp.          // 200
//                                                                                                    // 201
// Each of daily, weekly, and monthly contains:                                                       // 202
//   activeUsers: The number of unique users who have used a grain on the server in the time          // 203
//       interval. Only counts logged-in users.                                                       // 204
//   demoUsers: Demo users.                                                                           // 205
//   appDemoUsers: Users that came in through "app demo".                                             // 206
//   activeGrains: The number of unique grains that have been used in the time interval.              // 207
                                                                                                      // 208
DeleteStats = new Mongo.Collection("deleteStats");                                                    // 209
// Contains records of objects that were deleted, for stat-keeping purposes.                          // 210
//                                                                                                    // 211
// Each contains:                                                                                     // 212
//   type: "grain" or "user" or "demoUser" or "appDemoUser"                                           // 213
//   lastActive: Date of the user's or grain's last activity.                                         // 214
                                                                                                      // 215
FileTokens = new Mongo.Collection("fileTokens");                                                      // 216
// Tokens corresponding to backup files that are currently stored on the server. A user receives      // 217
// a token when they create a backup file (either by uploading it, or by backing up one of their      // 218
// grains) and may use the token to read the file (either to download it, or to restore a new         // 219
// grain from it).                                                                                    // 220
//                                                                                                    // 221
// Each contains:                                                                                     // 222
//   _id:       The unguessable token string.                                                         // 223
//   name:      Suggested filename.                                                                   // 224
//   timestamp: File creation time. Used to figure out when the token and file should be wiped.       // 225
                                                                                                      // 226
ApiTokens = new Mongo.Collection("apiTokens");                                                        // 227
// Access tokens for APIs exported by apps.                                                           // 228
//                                                                                                    // 229
// Originally API tokens were only used by external users through the HTTP API endpoint. However,     // 230
// now they are also used to implement SturdyRefs, not just held by external users, but also when     // 231
// an app holds a SturdyRef to another app within the same server. See the various `save()`,          // 232
// `restore()`, and `drop()` methods in `grain.capnp` (on `SansdtormApi`, `AppPersistent`, and        // 233
// `MainView`) -- the fields of type `Data` are API tokens.                                           // 234
//                                                                                                    // 235
// Each contains:                                                                                     // 236
//   _id:       A SHA-256 hash of the token.                                                          // 237
//   grainId:   The grain servicing this API. (Not present if the API isn't serviced by a grain.)     // 238
//   userId:    For UiView capabilities, this is the user for whom the view is attenuated. That       // 239
//              is, the UiView's newSession() method will intersect the requested permissions with    // 240
//              this user's permissions before forwarding on to the underlying app. If `userId` is    // 241
//              not present, then no user attenuation is applied, i.e. this is a raw UiView as        // 242
//              implemented by the app. (The `roleAssignment` field, below, may still apply.)         // 243
//              For non-UiView capabilities, `userId` is never present.                               // 244
//              Note that this is NOT the user against whom the `requiredPermissions` parameter of    // 245
//              `SandstormApi.restore()` is checked; that would be `owner.grain.introducerUser`.      // 246
//   userInfo:  *DEPRECATED* For API tokens created by the app through HackSessionContext, the        // 247
//              UserInfo struct that should be passed to `newSession()` when exercising this token,   // 248
//              in decoded (JS object) format. This is a temporary hack. `userId` is never present    // 249
//              when `userInfo` is present.                                                           // 250
//   roleAssignment: If this API token represents a UiView, this field contains a JSON-encoded        // 251
//              Grain.ViewSharingLink.RoleAssignment representing the permissions it carries. These   // 252
//              permissions will be intersected with those held by `userId` when the view is opened.  // 253
//   forSharing: If true, requests sent to the HTTP API endpoint with this token will be treated as   // 254
//              anonymous rather than as directly associated with `userId`. This has no effect on     // 255
//              the permissions granted.                                                              // 256
//   objectId:  If present, this token represents an arbitrary Cap'n Proto capability exported by     // 257
//              the app or its supervisor (whereas without this it strictly represents UiView).       // 258
//              sturdyRef is the JSON-encoded SupervisorObjectId (defined in `supervisor.capnp`).     // 259
//              Note that if the SupervisorObjectId contains an AppObjectId, that field is            // 260
//              treated as type AnyPointer, and so encoded as a raw Cap'n Proto message.              // 261
//   frontendRef: If present, this token actually refers to an object implemented by the front-end,   // 262
//              not a particular grain. (`grainId` and `userId` are not set.) This is an object       // 263
//              containing exactly one of the following fields:                                       // 264
//       notificationHandle: A `Handle` for an ongoing notification, as returned by                   // 265
//                           `NotificationTarget.addOngoing`. The value is an `_id` from the          // 266
//                           `Notifications` collection.                                              // 267
//       ipNetwork: An IpNetwork capability that is implemented by the frontend. Eventually, this     // 268
//                  will be moved out of the frontend and into the backend, but we'll migrate the     // 269
//                  database when that happens. This field contains the boolean true to signify that  // 270
//                  it has been set.                                                                  // 271
//       ipInterface: Ditto IpNetwork, except it's an IpInterface.                                    // 272
//   parentToken: If present, then this token represents exactly the capability represented by        // 273
//              the ApiToken with _id = parentToken, except possibly (if it is a UiView) attenuated   // 274
//              by `roleAssignment` (if present). To facilitate permissions computations, if the      // 275
//              capability is a UiView, then `grainId` is set to the backing grain and `userId` is    // 276
//              set to the user who shared the view. None of `userInfo`, `objectId`, or               // 277
//              `frontendRef` are present when `parentToken` is present.                              // 278
//   petname:   Human-readable label for this access token, useful for identifying tokens for         // 279
//              revocation. This should be displayed when visualizing incoming capabilities to        // 280
//              the grain identified by `grainId`.                                                    // 281
//   created:   Date when this token was created.                                                     // 282
//   revoked:   If true, then this sturdyref has been revoked and can no longer be restored. It may   // 283
//              become un-revoked in the future.                                                      // 284
//   expires:   Optional expiration Date. If undefined, the token does not expire.                    // 285
//   owner:     A `ApiTokenRefOwner` (defined in `supervisor.capnp`, stored as a JSON object)         // 286
//              as passed to the `save()` call that created this token. If not present, treat         // 287
//              as `webkey` (the default for `ApiTokenOwner`).                                        // 288
//   expiresIfUnused:                                                                                 // 289
//              Optional Date after which the token, if it has not been used yet, expires.            // 290
//              This field should be cleared on a token's first use.                                  // 291
//   requirements: List of conditions which must hold for this token to be considered valid.          // 292
//              Semantically, this list specifies the powers which were *used* to originally          // 293
//              create the token. If any condition in the list becomes untrue, then the token must    // 294
//              be considered revoked, and all live refs and sturdy refs obtained transitively        // 295
//              through it must also become revoked. Each item is the JSON serialization of the       // 296
//              `MembraneRequirement` structure defined in `supervisor.capnp`.                        // 297
//                                                                                                    // 298
// It is important to note that a token's owner and provider are independent from each other. To      // 299
// illustrate, here is an approximate definition of ApiToken in pseudo Cap'n Proto schema language:   // 300
//                                                                                                    // 301
// struct ApiToken {                                                                                  // 302
//   owner :ApiTokenOwner;                                                                            // 303
//   provider :union {                                                                                // 304
//     grain :group {                                                                                 // 305
//       grainId :Text;                                                                               // 306
//       union {                                                                                      // 307
//         uiView :group {                                                                            // 308
//           userId :Text;                                                                            // 309
//           roleAssignment :RoleAssignment;                                                          // 310
//           forSharing :Bool;                                                                        // 311
//         }                                                                                          // 312
//         objectId :SupervisorObjectId;                                                              // 313
//       }                                                                                            // 314
//     }                                                                                              // 315
//     frontendRef :union {                                                                           // 316
//        notificationHandle :Text;                                                                   // 317
//        ipNetwork :Bool;                                                                            // 318
//        ipInterface :Bool;                                                                          // 319
//     }                                                                                              // 320
//     child :group {                                                                                 // 321
//       parentToken :Text;                                                                           // 322
//       union {                                                                                      // 323
//         uiView :group {                                                                            // 324
//           grainId :Text;                                                                           // 325
//           userId :Text;                                                                            // 326
//           roleAssignment :RoleAssignment = (allAccess = ());                                       // 327
//         }                                                                                          // 328
//         other :Void;                                                                               // 329
//       }                                                                                            // 330
//     }                                                                                              // 331
//   }                                                                                                // 332
//   requirements: List(Supervisor.MembraneRequirement);                                              // 333
//   ...                                                                                              // 334
// }                                                                                                  // 335
                                                                                                      // 336
Notifications = new Mongo.Collection("notifications");                                                // 337
// Notifications for a user.                                                                          // 338
//                                                                                                    // 339
// Each contains:                                                                                     // 340
//   _id:          random                                                                             // 341
//   ongoing:      If present, this is an ongoing notification, and this field contains an            // 342
//                 ApiToken referencing the `OngoingNotification` capability.                         // 343
//   grainId:      The grain originating this notification, if any.                                   // 344
//   userId:       The user receiving the notification.                                               // 345
//   text:         The JSON-ified LocalizedText to display in the notification.                       // 346
//   isUnread:     Boolean indicating if this notification is unread.                                 // 347
//   timestamp:    Date when this notification was last updated                                       // 348
                                                                                                      // 349
StatsTokens = new Mongo.Collection("statsTokens");                                                    // 350
// Access tokens for the Stats collection                                                             // 351
//                                                                                                    // 352
// These tokens are used for accessing the ActivityStats collection remotely                          // 353
// (ie. from a dashboard webapp)                                                                      // 354
//                                                                                                    // 355
// Each contains:                                                                                     // 356
//   _id:       The token. At least 128 bits entropy (Random.id(22)).                                 // 357
                                                                                                      // 358
Misc = new Mongo.Collection("misc");                                                                  // 359
// Miscellaneous configuration and other settings                                                     // 360
//                                                                                                    // 361
// This table is currently only used for persisting BASE_URL from one session to the next,            // 362
// but in general any miscellaneous settings should go in here                                        // 363
//                                                                                                    // 364
// Each contains:                                                                                     // 365
//   _id:       The name of the setting. eg. "BASE_URL"                                               // 366
//   value:     The value of the setting.                                                             // 367
                                                                                                      // 368
Settings = new Mongo.Collection("settings");                                                          // 369
// Settings for this Sandstorm instance go here. They are configured through the adminSettings        // 370
// route. This collection differs from misc in that this collection is completely user controlled.    // 371
//                                                                                                    // 372
// Each contains:                                                                                     // 373
//   _id:       The name of the setting. eg. "MAIL_URL"                                               // 374
//   value:     The value of the setting.                                                             // 375
//   potentially other fields that are unique to the setting                                          // 376
                                                                                                      // 377
Migrations = new Mongo.Collection("migrations");                                                      // 378
// This table tracks which migrations we have applied to this instance.                               // 379
// It contains a single entry:                                                                        // 380
//   _id:       "migrations_applied"                                                                  // 381
//   value:     The number of migrations this instance has successfully completed.                    // 382
                                                                                                      // 383
StaticAssets = new Mongo.Collection("staticAssets");                                                  // 384
// Collection of static assets served up from the Sandstorm server's "static" host. We only           // 385
// support relatively small assets: under 1MB each.                                                   // 386
//                                                                                                    // 387
// Each contains:                                                                                     // 388
//   _id:       Random ID; will be used in the URL.                                                   // 389
//   hash:      A SHA-256 hash of the data, used to de-dupe.                                          // 390
//   mimeType:  MIME type of the asset, suitable for Content-Type header.                             // 391
//   encoding:  Either "gzip" or not present, suitable for Content-Encoding header.                   // 392
//   content:   The asset content (byte buffer).                                                      // 393
//   refcount:  Number of places where this asset's ID appears in the database. Since Mongo doesn't   // 394
//       have transactions, this needs to bias towards over-counting; a backup GC could be used       // 395
//       to catch leaked assets, although it's probably not a big deal in practice.                   // 396
                                                                                                      // 397
AssetUploadTokens = new Mongo.Collection("assetUploadTokens");                                        // 398
// Collection of tokens representing a single-use permission to upload an asset, such as a new        // 399
// profile picture.                                                                                   // 400
//                                                                                                    // 401
// Each contains:                                                                                     // 402
//   _id:       Random ID.                                                                            // 403
//   purpose:   Contains one of the following, indicating how the asset is to be used:                // 404
//       profilePicture: Indicates that the upload is a new profile picture. Contains fields:         // 405
//           userId: User whose picture shall be replaced.                                            // 406
//           identityId: Which of the user's identities shall be updated.                             // 407
//   expires:   Time when this token will go away if unused.                                          // 408
                                                                                                      // 409
Plans = new Mongo.Collection("plans");                                                                // 410
// Subscription plans, which determine quota.                                                         // 411
//                                                                                                    // 412
// Each contains:                                                                                     // 413
//   _id: Plan ID, usually a short string like "free", "standard", "large", "mega", ...               // 414
//   storage: Number of bytes this user is allowed to store.                                          // 415
//   compute: Number of kilobyte-RAM-seconds this user is allowed to consume.                         // 416
//   computeLabel: Label to display to the user describing this plan's compute units.                 // 417
//   grains: Total number of grains this user can create (often `Infinity`).                          // 418
//   price: Price per month in US cents.                                                              // 419
                                                                                                      // 420
if (Meteor.isServer) {                                                                                // 421
  Meteor.publish("credentials", function () {                                                         // 422
    // Data needed for isSignedUp() and isAdmin() to work.                                            // 423
                                                                                                      // 424
    if (this.userId) {                                                                                // 425
      return [                                                                                        // 426
        Meteor.users.find({_id: this.userId},                                                         // 427
            {fields: {signupKey: 1, isAdmin: 1, expires: 1, storageUsage: 1,                          // 428
                      plan: 1, hasCompletedSignup: 1}}),                                              // 429
        Plans.find()                                                                                  // 430
      ];                                                                                              // 431
    } else {                                                                                          // 432
      return [];                                                                                      // 433
    }                                                                                                 // 434
  });                                                                                                 // 435
}                                                                                                     // 436
                                                                                                      // 437
isDemoUser = function() {                                                                             // 438
  // Returns true if this is a demo user.                                                             // 439
                                                                                                      // 440
  var user = Meteor.user();                                                                           // 441
  if (user && user.expires) {                                                                         // 442
    return true;                                                                                      // 443
  } else {                                                                                            // 444
    return false;                                                                                     // 445
  }                                                                                                   // 446
}                                                                                                     // 447
                                                                                                      // 448
isSignedUp = function() {                                                                             // 449
  // Returns true if the user has presented an invite key.                                            // 450
                                                                                                      // 451
  var user = Meteor.user();                                                                           // 452
                                                                                                      // 453
  if (!user) return false;  // not signed in                                                          // 454
                                                                                                      // 455
  if (user.expires) return false;  // demo user.                                                      // 456
                                                                                                      // 457
  if (Meteor.settings.public.allowUninvited) return true;  // all accounts qualify                    // 458
                                                                                                      // 459
  if (user.signupKey) return true;  // user is invited                                                // 460
                                                                                                      // 461
  return false;                                                                                       // 462
}                                                                                                     // 463
                                                                                                      // 464
isSignedUpOrDemo = function () {                                                                      // 465
  var user = Meteor.user();                                                                           // 466
                                                                                                      // 467
  if (!user) return false;  // not signed in                                                          // 468
                                                                                                      // 469
  if (user.expires) return true;  // demo user.                                                       // 470
                                                                                                      // 471
  if (Meteor.settings.public.allowUninvited) return true;  // all accounts qualify                    // 472
                                                                                                      // 473
  if (user.signupKey) return true;  // user is invited                                                // 474
                                                                                                      // 475
  return false;                                                                                       // 476
}                                                                                                     // 477
                                                                                                      // 478
isUserOverQuota = function (user) {                                                                   // 479
  // Return false if user has quota space remaining, true if it is full. When this returns true,      // 480
  // we will not allow the user to create new grains, though they may be able to open existing ones   // 481
  // which may still increase their storage usage.                                                    // 482
  //                                                                                                  // 483
  // (Actually returns a string which can be fed into `billingPrompt` as the reason.)                 // 484
                                                                                                      // 485
  if (!Meteor.settings.public.quotaEnabled || user.isAdmin) return false;                             // 486
                                                                                                      // 487
  var plan = Plans.findOne(user.plan || "free");                                                      // 488
                                                                                                      // 489
  if (plan.grains < Infinity) {                                                                       // 490
    var count = Grains.find({userId: user._id}, {fields: {}, limit: plan.grains}).count();            // 491
    if (count >= plan.grains) return "outOfGrains";                                                   // 492
  }                                                                                                   // 493
                                                                                                      // 494
  return plan && user.storageUsage && user.storageUsage >= plan.storage && "outOfStorage";            // 495
}                                                                                                     // 496
                                                                                                      // 497
isUserExcessivelyOverQuota = function (user) {                                                        // 498
  // Return true if user is so far over quota that we should prevent their existing grains from       // 499
  // running at all.                                                                                  // 500
  //                                                                                                  // 501
  // (Actually returns a string which can be fed into `billingPrompt` as the reason.)                 // 502
                                                                                                      // 503
  if (!Meteor.settings.public.quotaEnabled || user.isAdmin) return false;                             // 504
                                                                                                      // 505
  var plan = Plans.findOne(user.plan || "free");                                                      // 506
                                                                                                      // 507
  if (plan.grains < Infinity) {                                                                       // 508
    var count = Grains.find({userId: user._id}, {fields: {}, limit: plan.grains * 2}).count();        // 509
    if (count >= plan.grains * 2) return "outOfGrains";                                               // 510
  }                                                                                                   // 511
                                                                                                      // 512
  return plan && user.storageUsage && user.storageUsage >= plan.storage * 1.2 && "outOfStorage";      // 513
}                                                                                                     // 514
                                                                                                      // 515
isAdmin = function() {                                                                                // 516
  // Returns true if the user is the administrator.                                                   // 517
                                                                                                      // 518
  var user = Meteor.user();                                                                           // 519
  if (user && user.isAdmin) {                                                                         // 520
    return true;                                                                                      // 521
  } else {                                                                                            // 522
    return false;                                                                                     // 523
  }                                                                                                   // 524
}                                                                                                     // 525
                                                                                                      // 526
isAdminById = function(id) {                                                                          // 527
  // Returns true if the user's id is the administrator.                                              // 528
                                                                                                      // 529
  var user = Meteor.users.findOne({_id: id}, {fields: {isAdmin: 1}})                                  // 530
  if (user && user.isAdmin) {                                                                         // 531
    return true;                                                                                      // 532
  } else {                                                                                            // 533
    return false;                                                                                     // 534
  }                                                                                                   // 535
}                                                                                                     // 536
                                                                                                      // 537
findAdminUserForToken = function (token) {                                                            // 538
  if (!token.requirements) {                                                                          // 539
    return;                                                                                           // 540
  }                                                                                                   // 541
  var requirements = token.requirements.filter(function (requirement) {                               // 542
    return "userIsAdmin" in requirement;                                                              // 543
  });                                                                                                 // 544
                                                                                                      // 545
  if (requirements.length > 1) {                                                                      // 546
    return;                                                                                           // 547
  }                                                                                                   // 548
  if (requirements.length === 0) {                                                                    // 549
    return;                                                                                           // 550
  }                                                                                                   // 551
  return requirements[0].userIsAdmin;                                                                 // 552
};                                                                                                    // 553
                                                                                                      // 554
var wildcardHost = Meteor.settings.public.wildcardHost.toLowerCase().split("*");                      // 555
                                                                                                      // 556
if (wildcardHost.length != 2) {                                                                       // 557
  throw new Error("Wildcard host must contain exactly one asterisk.");                                // 558
}                                                                                                     // 559
                                                                                                      // 560
matchWildcardHost = function(host) {                                                                  // 561
  // See if the hostname is a member of our wildcard. If so, extract the ID.                          // 562
                                                                                                      // 563
  var prefix = wildcardHost[0];                                                                       // 564
  var suffix = wildcardHost[1];                                                                       // 565
  if (host.lastIndexOf(prefix, 0) >= 0 &&                                                             // 566
      host.indexOf(suffix, -suffix.length) >= 0 &&                                                    // 567
      host.length >= prefix.length + suffix.length) {                                                 // 568
    var id = host.slice(prefix.length, -suffix.length);                                               // 569
    if (id.match(/^[a-z0-9]*$/)) {                                                                    // 570
      return id;                                                                                      // 571
    }                                                                                                 // 572
  }                                                                                                   // 573
                                                                                                      // 574
  return null;                                                                                        // 575
}                                                                                                     // 576
                                                                                                      // 577
makeWildcardHost = function (id) {                                                                    // 578
  return wildcardHost[0] + id + wildcardHost[1];                                                      // 579
}                                                                                                     // 580
                                                                                                      // 581
if (Meteor.isServer) {                                                                                // 582
  var Url = Npm.require("url");                                                                       // 583
  getWildcardOrigin = function () {                                                                   // 584
    // The wildcard URL can be something like "foo-*-bar.example.com", but sometimes when we're       // 585
    // trying to specify a pattern matching hostnames (say, a Content-Security-Policy directive),     // 586
    // an astrisk is only allowed as the first character and must be followed by a period. So we need // 587
    // "*.example.com" instead -- which matches more than we actually want, but is the best we can    // 588
    // really do. We also add the protocol to the front (again, that's what CSP wants).               // 589
                                                                                                      // 590
    // TODO(cleanup): `protocol` is computed in other files, like proxy.js. Put it somewhere common.  // 591
    var protocol = Url.parse(process.env.ROOT_URL).protocol;                                          // 592
                                                                                                      // 593
    var dotPos = wildcardHost[1].indexOf(".");                                                        // 594
    if (dotPos < 0) {                                                                                 // 595
      return protocol + "//*";                                                                        // 596
    } else {                                                                                          // 597
      return protocol + "//*" + wildcardHost[1].slice(dotPos);                                        // 598
    }                                                                                                 // 599
  }                                                                                                   // 600
}                                                                                                     // 601
                                                                                                      // 602
allowDevAccounts = function () {                                                                      // 603
  var setting = Settings.findOne({_id: "devAccounts"});                                               // 604
  if (setting) {                                                                                      // 605
    return setting.value;                                                                             // 606
  } else {                                                                                            // 607
    return Meteor.settings && Meteor.settings.public &&                                               // 608
           Meteor.settings.public.allowDevAccounts;                                                   // 609
  }                                                                                                   // 610
};                                                                                                    // 611
                                                                                                      // 612
roleAssignmentPattern = {                                                                             // 613
  none : Match.Optional(null),                                                                        // 614
  allAccess: Match.Optional(null),                                                                    // 615
  roleId: Match.Optional(Match.Integer),                                                              // 616
  addPermissions: Match.Optional([Boolean]),                                                          // 617
  removePermissions: Match.Optional([Boolean]),                                                       // 618
};                                                                                                    // 619
                                                                                                      // 620
SandstormDb = function () {                                                                           // 621
  this.collections = {                                                                                // 622
    // Direct access to underlying collections. DEPRECATED.                                           // 623
    //                                                                                                // 624
    // TODO(cleanup): Over time, we will provide methods covering each supported query and remove     // 625
    //   direct access to the collections.                                                            // 626
                                                                                                      // 627
    packages: Packages,                                                                               // 628
    devApps: DevApps,                                                                                 // 629
    userActions: UserActions,                                                                         // 630
    grains: Grains,                                                                                   // 631
    contacts: Contacts,                                                                               // 632
    sessions: Sessions,                                                                               // 633
    signupKeys: SignupKeys,                                                                           // 634
    activityStats: ActivityStats,                                                                     // 635
    deleteStats: DeleteStats,                                                                         // 636
    fileTokens: FileTokens,                                                                           // 637
    apiTokens: ApiTokens,                                                                             // 638
    notifications: Notifications,                                                                     // 639
    statsTokens: StatsTokens,                                                                         // 640
    misc: Misc,                                                                                       // 641
    settings: Settings,                                                                               // 642
                                                                                                      // 643
    // Intentionally omitted:                                                                         // 644
    // - Migrations, since it's used only within this package.                                        // 645
    // - RoleAssignments, since it is deprecated and only used by the migration that eliminated it.   // 646
  };                                                                                                  // 647
};                                                                                                    // 648
                                                                                                      // 649
// TODO(cleanup): These methods should not be defined freestanding and should use collection          // 650
//   objects created in SandstormDb's constructor rather than globals.                                // 651
                                                                                                      // 652
_.extend(SandstormDb.prototype, {                                                                     // 653
  isDemoUser: isDemoUser,                                                                             // 654
  isSignedUp: isSignedUp,                                                                             // 655
  isSignedUpOrDemo: isSignedUpOrDemo,                                                                 // 656
  isUserOverQuota: isUserOverQuota,                                                                   // 657
  isUserExcessivelyOverQuota: isUserExcessivelyOverQuota,                                             // 658
  isAdmin: isAdmin,                                                                                   // 659
  isAdminById: isAdminById,                                                                           // 660
  findAdminUserForToken: findAdminUserForToken,                                                       // 661
  matchWildcardHost: matchWildcardHost,                                                               // 662
  makeWildcardHost: makeWildcardHost,                                                                 // 663
  allowDevAccounts: allowDevAccounts,                                                                 // 664
  roleAssignmentPattern: roleAssignmentPattern,                                                       // 665
});                                                                                                   // 666
                                                                                                      // 667
if (Meteor.isServer) {                                                                                // 668
  SandstormDb.prototype.getWildcardOrigin = getWildcardOrigin;                                        // 669
}                                                                                                     // 670
                                                                                                      // 671
// =======================================================================================            // 672
// Below this point are newly-written or refactored functions.                                        // 673
                                                                                                      // 674
_.extend(SandstormDb.prototype, {                                                                     // 675
  // TODO(cleanup): These methods shouldn't take a raw mongo queries or "aggregations" as inputs.     // 676
  //   We want methods corresponding to each kind of query that is performed in practice. Letting     // 677
  //   the caller pass an arbitrary query and aggregations is problematic because:                    // 678
  //   - We can't type-check them to prevent Mongo injections.                                        // 679
  //   - It defeats the purpose of clearly seeing what kinds of queries we need to support, and       // 680
  //     therefore what kind of indexes we might need.                                                // 681
  //   - It doesn't make it any easier for us to substitute a non-Mongo database in the future.       // 682
  //   - If the caller is just going to pass in a match-all query, we may be forcing Mongo into       // 683
  //     a slower path by using $and.                                                                 // 684
  userGrains: function userGrains (user, query, aggregations) {                                       // 685
    var filteredQuery = { $and: [ { userId: user}, query ] };                                         // 686
    return this.collections.grains.find(filteredQuery, aggregations);                                 // 687
  },                                                                                                  // 688
                                                                                                      // 689
  currentUserGrains: function currentUserGrains (query, aggregations) {                               // 690
    return this.userGrains(Meteor.userId(), query, aggregations);                                     // 691
  },                                                                                                  // 692
                                                                                                      // 693
  userApiTokens: function userApiTokens (user) {                                                      // 694
    return this.collections.apiTokens.find({'owner.user.userId': user});                              // 695
  },                                                                                                  // 696
                                                                                                      // 697
  currentUserApiTokens: function currentUserApiTokens () {                                            // 698
    return this.userApiTokens(Meteor.userId());                                                       // 699
  },                                                                                                  // 700
                                                                                                      // 701
  userActions: function userActions (user, query, aggregations) {                                     // 702
    var filteredQuery = { $and: [ {userId: user}, query ] };                                          // 703
    return this.collections.userActions.find(filteredQuery, aggregations);                            // 704
  },                                                                                                  // 705
                                                                                                      // 706
  currentUserActions: function currentUserActions (query, aggregations) {                             // 707
    return this.userActions(Meteor.userId(), query, aggregations);                                    // 708
  },                                                                                                  // 709
                                                                                                      // 710
  getPlan: function (id) {                                                                            // 711
    check(id, String);                                                                                // 712
    var plan = Plans.findOne(id);                                                                     // 713
    if (!plan) {                                                                                      // 714
      throw new Error("no such plan: " + id);                                                         // 715
    }                                                                                                 // 716
    return plan;                                                                                      // 717
  },                                                                                                  // 718
                                                                                                      // 719
  listPlans: function () {                                                                            // 720
    return Plans.find({}, {sort: {price: 1}});                                                        // 721
  },                                                                                                  // 722
                                                                                                      // 723
  getMyPlan: function () {                                                                            // 724
    var user = Meteor.user();                                                                         // 725
    return user && Plans.findOne(user.plan || "free");                                                // 726
  },                                                                                                  // 727
                                                                                                      // 728
  getMyUsage: function (user) {                                                                       // 729
    user = user || Meteor.user();                                                                     // 730
    if (user && (Meteor.isServer || user.pseudoUsage)) {                                              // 731
      if (Meteor.isClient) {                                                                          // 732
        // Filled by pseudo-subscription to "getMyUsage". WARNING: The subscription is currenly       // 733
        // not reactive.                                                                              // 734
        return user.pseudoUsage;                                                                      // 735
      } else {                                                                                        // 736
        return {                                                                                      // 737
          grains: Grains.find({userId: user._id}).count(),                                            // 738
          storage: user.storageUsage || 0,                                                            // 739
          compute: 0   // not tracked yet                                                             // 740
        };                                                                                            // 741
      }                                                                                               // 742
    } else {                                                                                          // 743
      return {grains: 0, storage: 0, compute: 0};                                                     // 744
    }                                                                                                 // 745
  },                                                                                                  // 746
                                                                                                      // 747
  isUninvitedFreeUser: function () {                                                                  // 748
    if (!Meteor.settings.public.allowUninvited) return false;                                         // 749
                                                                                                      // 750
    var user = Meteor.user();                                                                         // 751
    return user && !user.expires && (!user.plan || user.plan === "free");                             // 752
  },                                                                                                  // 753
                                                                                                      // 754
  getSetting: function (name) {                                                                       // 755
    var setting = Settings.findOne(name);                                                             // 756
    return setting && setting.value;                                                                  // 757
  },                                                                                                  // 758
});                                                                                                   // 759
                                                                                                      // 760
if (Meteor.isServer) {                                                                                // 761
  var Crypto = Npm.require("crypto");                                                                 // 762
  var ContentType = Npm.require("content-type");                                                      // 763
  var Zlib = Npm.require("zlib");                                                                     // 764
                                                                                                      // 765
  var replicaNumber = Meteor.settings.replicaNumber || 0;                                             // 766
                                                                                                      // 767
  var computeStagger = function (n) {                                                                 // 768
    // Compute a fraction in the range [0, 1) such that, for any natural number k, the values         // 769
    // of computeStagger(n) for all n in [1, 2^k) are uniformly distributed between 0 and 1.          // 770
    // The sequence looks like:                                                                       // 771
    //   0, 1/2, 1/4, 3/4, 1/8, 3/8, 5/8, 7/8, 1/16, ...                                              // 772
    //                                                                                                // 773
    // We use this to determine how we'll stagger periodic events performed by this replica.          // 774
    // Notice that this allows us to compute a stagger which is independent of the number of          // 775
    // front-end replicas present; we can add more replicas to the end without affecting how the      // 776
    // earlier ones schedule their events.                                                            // 777
    var denom = 1;                                                                                    // 778
    while (denom <= n) denom <<= 1;                                                                   // 779
    var num = n * 2 - denom + 1;                                                                      // 780
    return num / denom;                                                                               // 781
  }                                                                                                   // 782
                                                                                                      // 783
  var stagger = computeStagger(replicaNumber);                                                        // 784
                                                                                                      // 785
  SandstormDb.periodicCleanup = function (intervalMs, callback) {                                     // 786
    // Register a database cleanup function than should run periodically, roughly once every          // 787
    // interval of the given length.                                                                  // 788
    //                                                                                                // 789
    // In a blackrock deployment with multiple front-ends, the frequency of the cleanup will be       // 790
    // scaled appropriately on the assumption that more data is being generated demanding more        // 791
    // frequent cleanups.                                                                             // 792
                                                                                                      // 793
    check(intervalMs, Number);                                                                        // 794
    check(callback, Function);                                                                        // 795
                                                                                                      // 796
    if (intervalMs < 120000) {                                                                        // 797
      throw new Error("less than 2-minute cleanup interval seems too fast; " +                        // 798
                      "are you using the right units?");                                              // 799
    }                                                                                                 // 800
                                                                                                      // 801
    // Schedule first cleanup to happen at the next intervalMs interval from the epoch, so that       // 802
    // the schedule is independent of the exact startup time.                                         // 803
    var first = intervalMs - Date.now() % intervalMs;                                                 // 804
                                                                                                      // 805
    // Stagger cleanups across replicas so that we don't have all replicas trying to clean the        // 806
    // same data at the same time.                                                                    // 807
    first += Math.floor(intervalMs * computeStagger(replicaNumber));                                  // 808
                                                                                                      // 809
    // If the stagger put us more than an interval away from now, back up.                            // 810
    if (first > intervalMs) first -= intervalMs;                                                      // 811
                                                                                                      // 812
    Meteor.setTimeout(function () {                                                                   // 813
      callback();                                                                                     // 814
      Meteor.setInterval(callback, intervalMs);                                                       // 815
    }, first);                                                                                        // 816
  }                                                                                                   // 817
                                                                                                      // 818
  // TODO(cleanup): Node 0.12 has a `gzipSync` but 0.10 (which Meteor still uses) does not.           // 819
  var gzipSync = Meteor.wrapAsync(Zlib.gzip, Zlib);                                                   // 820
                                                                                                      // 821
  var BufferSmallerThan = function (limit) {                                                          // 822
    return Match.Where(function (buf) {                                                               // 823
      check(buf, Buffer);                                                                             // 824
      return buf.length < limit;                                                                      // 825
    });                                                                                               // 826
  }                                                                                                   // 827
                                                                                                      // 828
  var DatabaseId = Match.Where(function (s) {                                                         // 829
    check(s, String);                                                                                 // 830
    return !!s.match(/^[a-zA-Z0-9_]+$/);                                                              // 831
  });                                                                                                 // 832
                                                                                                      // 833
  addStaticAsset = function (metadata, content) {                                                     // 834
    // Add a new static asset to the database. If `content` is a string rather than a buffer, it      // 835
    // will be automatically gzipped before storage; do not specify metadata.encoding in this case.   // 836
                                                                                                      // 837
    if (typeof content === "string" && !metadata.encoding) {                                          // 838
      content = gzipSync(new Buffer(content, "utf8"));                                                // 839
      metadata.encoding = "gzip";                                                                     // 840
    }                                                                                                 // 841
                                                                                                      // 842
    check(metadata, {                                                                                 // 843
      mimeType: String,                                                                               // 844
      encoding: Match.Optional("gzip")                                                                // 845
    });                                                                                               // 846
    check(content, BufferSmallerThan(1 << 20));                                                       // 847
                                                                                                      // 848
    // Validate content type.                                                                         // 849
    metadata.mimeType = ContentType.format(ContentType.parse(metadata.mimeType));                     // 850
                                                                                                      // 851
    var hasher = Crypto.createHash("sha256");                                                         // 852
    hasher.update(metadata.mimeType + "\n" + metadata.encoding + "\n", "utf8");                       // 853
    hasher.update(content);                                                                           // 854
    var hash = hasher.digest("base64");                                                               // 855
                                                                                                      // 856
    var existing = StaticAssets.findAndModify({                                                       // 857
      query: {hash: hash, refcount: {$gte: 1}},                                                       // 858
      update: {$inc: {refcount: 1}},                                                                  // 859
      fields: {_id: 1, refcount: 1},                                                                  // 860
    });                                                                                               // 861
    if (existing) {                                                                                   // 862
      return existing._id;                                                                            // 863
    }                                                                                                 // 864
                                                                                                      // 865
    return StaticAssets.insert(_.extend({                                                             // 866
      hash: hash,                                                                                     // 867
      content: content,                                                                               // 868
      refcount: 1                                                                                     // 869
    }, metadata));                                                                                    // 870
  }                                                                                                   // 871
                                                                                                      // 872
  SandstormDb.prototype.addStaticAsset = addStaticAsset;                                              // 873
                                                                                                      // 874
  SandstormDb.prototype.refStaticAsset = function (id) {                                              // 875
    // Increment the refcount on an existing static asset.                                            // 876
    //                                                                                                // 877
    // You must call this BEFORE adding the new reference to the DB, in case of failure between       // 878
    // the two calls. (This way, the failure case is a storage leak, which is probably not a big      // 879
    // deal and can be fixed by GC, rather than a mysteriously missing asset.)                        // 880
                                                                                                      // 881
    check(id, String);                                                                                // 882
                                                                                                      // 883
    var existing = StaticAssets.findAndModify({                                                       // 884
      query: {hash: hash},                                                                            // 885
      update: {$inc: {refcount: 1}},                                                                  // 886
      fields: {_id: 1, refcount: 1},                                                                  // 887
    });                                                                                               // 888
    if (!existing) {                                                                                  // 889
      throw new Error("refStaticAsset() called on asset that doesn't exist");                         // 890
    }                                                                                                 // 891
  }                                                                                                   // 892
                                                                                                      // 893
  SandstormDb.prototype.unrefStaticAsset = function (id) {                                            // 894
    // Decrement refcount on a static asset and delete if it has reached zero.                        // 895
    //                                                                                                // 896
    // You must call this AFTER removing the reference from the DB, in case of failure between        // 897
    // the two calls. (This way, the failure case is a storage leak, which is probably not a big      // 898
    // deal and can be fixed by GC, rather than a mysteriously missing asset.)                        // 899
                                                                                                      // 900
    check(id, String);                                                                                // 901
                                                                                                      // 902
    var existing = StaticAssets.findAndModify({                                                       // 903
      query: {_id: id},                                                                               // 904
      update: {$inc: {refcount: -1}},                                                                 // 905
      fields: {_id: 1, refcount: 1},                                                                  // 906
      new: true,                                                                                      // 907
    });                                                                                               // 908
    if (!existing) {                                                                                  // 909
      console.error(new Error("unrefStaticAsset() called on asset that doesn't exist").stack);        // 910
    } else if (existing.refcount <= 0) {                                                              // 911
      StaticAssets.remove({_id: existing._id});                                                       // 912
    }                                                                                                 // 913
  }                                                                                                   // 914
                                                                                                      // 915
  SandstormDb.prototype.getStaticAsset = function (id) {                                              // 916
    // Get a static asset's mimeType, encoding, and raw content.                                      // 917
                                                                                                      // 918
    check(id, String);                                                                                // 919
                                                                                                      // 920
    var asset = StaticAssets.findOne(id, {fields: {_id: 0, mimeType: 1, encoding: 1, content: 1}});   // 921
    if (asset) {                                                                                      // 922
      // TODO(perf): Mongo converts buffers to something else. Figure out a way to avoid a copy       // 923
      //   here.                                                                                      // 924
      asset.content = new Buffer(asset.content);                                                      // 925
    }                                                                                                 // 926
    return asset;                                                                                     // 927
  }                                                                                                   // 928
                                                                                                      // 929
  SandstormDb.prototype.newAssetUpload = function (purpose) {                                         // 930
    check(purpose, {profilePicture: {userId: DatabaseId, identityId: Match.Optional(DatabaseId)}});   // 931
                                                                                                      // 932
    return AssetUploadTokens.insert({                                                                 // 933
      purpose: purpose,                                                                               // 934
      expires: new Date(Date.now() + 300000),  // in 5 minutes                                        // 935
    });                                                                                               // 936
  }                                                                                                   // 937
                                                                                                      // 938
  SandstormDb.prototype.fulfillAssetUpload = function (id) {                                          // 939
    // Indicates that the given asset upload has completed. It will be removed and its purpose        // 940
    // returned. If no matching upload exists, returns undefined.                                     // 941
                                                                                                      // 942
    check(id, String);                                                                                // 943
                                                                                                      // 944
    var upload = AssetUploadTokens.findAndModify({                                                    // 945
      query: {_id: id},                                                                               // 946
      remove: true                                                                                    // 947
    });                                                                                               // 948
                                                                                                      // 949
    if (upload.expires.valueOf() < Date.now()) {                                                      // 950
      return undefined;  // already expired                                                           // 951
    } else {                                                                                          // 952
      return upload.purpose;                                                                          // 953
    }                                                                                                 // 954
  }                                                                                                   // 955
                                                                                                      // 956
  function cleanupExpiredAssetUploads() {                                                             // 957
    AssetUploadTokens.remove({expires: {$lt: Date.now()}});                                           // 958
  }                                                                                                   // 959
                                                                                                      // 960
  // Cleanup tokens every hour.                                                                       // 961
  SandstormDb.periodicCleanup(3600000, cleanupExpiredAssetUploads);                                   // 962
                                                                                                      // 963
  var packageCache = {};                                                                              // 964
  // Package info is immutable. Let's cache to save on mongo queries.                                 // 965
                                                                                                      // 966
  SandstormDb.prototype.getPackage = function (packageId) {                                           // 967
    // Get the given package record. Since package info is immutable, cache the data in the server    // 968
    // to reduce mongo query overhead, since it turns out we have to fetch specific packages a        // 969
    // lot.                                                                                           // 970
                                                                                                      // 971
    if (packageId in packageCache) {                                                                  // 972
      return packageCache[packageId];                                                                 // 973
    }                                                                                                 // 974
                                                                                                      // 975
    var package = Packages.findOne(packageId);                                                        // 976
    if (package && package.status === "ready") {                                                      // 977
      packageCache[packageId] = package;                                                              // 978
    }                                                                                                 // 979
    return package;                                                                                   // 980
  }                                                                                                   // 981
}                                                                                                     // 982
                                                                                                      // 983
////////////////////////////////////////////////////////////////////////////////////////////////////////

}).call(this);






(function () {

////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                    //
// packages/sandstorm-db/profile.js                                                                   //
//                                                                                                    //
////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                      //
// Sandstorm - Personal Cloud Sandbox                                                                 // 1
// Copyright (c) 2015 Sandstorm Development Group, Inc. and contributors                              // 2
// All rights reserved.                                                                               // 3
//                                                                                                    // 4
// Licensed under the Apache License, Version 2.0 (the "License");                                    // 5
// you may not use this file except in compliance with the License.                                   // 6
// You may obtain a copy of the License at                                                            // 7
//                                                                                                    // 8
//   http://www.apache.org/licenses/LICENSE-2.0                                                       // 9
//                                                                                                    // 10
// Unless required by applicable law or agreed to in writing, software                                // 11
// distributed under the License is distributed on an "AS IS" BASIS,                                  // 12
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                           // 13
// See the License for the specific language governing permissions and                                // 14
// limitations under the License.                                                                     // 15
                                                                                                      // 16
var makeIdenticon;                                                                                    // 17
var httpProtocol;                                                                                     // 18
var sha256;                                                                                           // 19
                                                                                                      // 20
if (Meteor.isServer) {                                                                                // 21
  Meteor.publish("accountIdentities", function () {                                                   // 22
    if (!Meteor.userId) return [];                                                                    // 23
                                                                                                      // 24
    return [                                                                                          // 25
      Meteor.users.find(Meteor.userId,                                                                // 26
        {fields: {                                                                                    // 27
          "profile":1,                                                                                // 28
          "devName":1,                                                                                // 29
          "expires":1,                                                                                // 30
                                                                                                      // 31
          "services.google.id":1,                                                                     // 32
          "services.google.email":1,                                                                  // 33
          "services.google.verified_email":1,                                                         // 34
          "services.google.name":1,                                                                   // 35
          "services.google.picture":1,                                                                // 36
          "services.google.gender":1,                                                                 // 37
                                                                                                      // 38
          "services.github.id":1,                                                                     // 39
          "services.github.email":1,                                                                  // 40
          "services.github.username":1,                                                               // 41
                                                                                                      // 42
          "services.emailToken.email":1                                                               // 43
        }})                                                                                           // 44
    ];                                                                                                // 45
  });                                                                                                 // 46
                                                                                                      // 47
  makeIdenticon = function (id) {                                                                     // 48
    // We only make identicons client-side.                                                           // 49
    return undefined;                                                                                 // 50
  }                                                                                                   // 51
                                                                                                      // 52
  var Url = Npm.require("url");                                                                       // 53
  httpProtocol = Url.parse(process.env.ROOT_URL).protocol;                                            // 54
                                                                                                      // 55
  var Crypto = Npm.require("crypto");                                                                 // 56
  sha256 = function (data) {                                                                          // 57
    return Crypto.createHash("sha256").update(data).digest("hex");                                    // 58
  };                                                                                                  // 59
} else {                                                                                              // 60
  var identiconCache = {};                                                                            // 61
                                                                                                      // 62
  makeIdenticon = function (id) {                                                                     // 63
    // Given a cryptographic hash as input, generate an identicon. We always use the user's           // 64
    // identity ID as the input hash (the same thing passed to apps as the user ID).                  // 65
                                                                                                      // 66
    // Pass to identicon.js exactly the hash that we'd pass to apps as X-Sandstorm-User-Id, so        // 67
    // that apps can themselves use identicon.js to produce consistent identicons. As it turns out    // 68
    // identicon.js doesn't use the second half of the hash even if we provide it, but slice it       // 69
    // anyway to be safe.                                                                             // 70
    id = id.slice(0, 32);                                                                             // 71
                                                                                                      // 72
    if (id in identiconCache) {                                                                       // 73
      return identiconCache[id];                                                                      // 74
    }                                                                                                 // 75
                                                                                                      // 76
    // Unfortunately, Github's algorithm uses MD5. Whatever, we don't expect these to be secure.      // 77
    var data = new Identicon(id, 512).toString();                                                     // 78
    var result = "data:image/png;base64," + data;                                                     // 79
    identiconCache[id] = result;                                                                      // 80
    return result;                                                                                    // 81
  }                                                                                                   // 82
                                                                                                      // 83
  httpProtocol = window.location.protocol;                                                            // 84
                                                                                                      // 85
  sha256 = function (data) {                                                                          // 86
    return CryptoJS.SHA256(data).toString();                                                          // 87
  };                                                                                                  // 88
}                                                                                                     // 89
                                                                                                      // 90
var GENDERS = {male: "male", female: "female", neutral: "neutral", robot: "robot"};                   // 91
                                                                                                      // 92
function staticAssetUrl(id, staticHost) {                                                             // 93
  if (id) {                                                                                           // 94
    return staticHost + "/" + id;                                                                     // 95
  } else {                                                                                            // 96
    return undefined;                                                                                 // 97
  }                                                                                                   // 98
}                                                                                                     // 99
                                                                                                      // 100
function filterHandle(handle) {                                                                       // 101
  // Convert disallowed letters into underscores, but no more than one underscore in a row.           // 102
  return handle && handle.toLowerCase().split(/[^a-z0-9_]+/g).join("_");                              // 103
}                                                                                                     // 104
                                                                                                      // 105
function emailToHandle(email) {                                                                       // 106
  // Turn an email address into a handle.                                                             // 107
                                                                                                      // 108
  if (!email) return undefined;                                                                       // 109
                                                                                                      // 110
  // Use the stuff before the @. Ignore stuff after '+' because it's commonly used for filters        // 111
  // on the same account.                                                                             // 112
  var parts = email.split("@");                                                                       // 113
  var base = filterHandle(parts[0].split("+")[0]);                                                    // 114
                                                                                                      // 115
  var domain = (parts[1]||"").split(".");                                                             // 116
  if (domain[domain.length - 1] === "name") {                                                         // 117
    // Oh, a .name domain. Let's use                                                                  // 118
    base = filterHandle(domain.slice(0, domain.length - 1).join("."));                                // 119
  } else if (_.contains(["me", "self", "contact", "admin", "administrator", "root", "info",           // 120
                         "sandstorm", "sandstormio", "inbox", "indiegogo", "mail", "email"],          // 121
                         base)) {                                                                     // 122
    // This is probably an address at a vanity domain. Use the domain itself as the handle.           // 123
    base = filterHandle(domain[0]);                                                                   // 124
  }                                                                                                   // 125
                                                                                                      // 126
  return filterHandle(base);                                                                          // 127
}                                                                                                     // 128
                                                                                                      // 129
function identityId(service, id) {                                                                    // 130
  return sha256(service + ":" + id);                                                                  // 131
}                                                                                                     // 132
                                                                                                      // 133
function googleIdentity(user, staticHost) {                                                           // 134
  var google = (user.services && user.services.google) || {};                                         // 135
  var profile = user.profile || {};                                                                   // 136
  var id = identityId("google", google.id);                                                           // 137
                                                                                                      // 138
  return {                                                                                            // 139
    service: "google",                                                                                // 140
    id: id,                                                                                           // 141
    name: profile.name || google.name || "Name Unknown",                                              // 142
    email: profile.email || google.email,                                                             // 143
    handle: profile.handle || emailToHandle(google.email) ||                                          // 144
            filterHandle(profile.name || google.name) || "unknown",                                   // 145
    picture: staticAssetUrl(profile.picture, staticHost) || makeIdenticon(id),                        // 146
    pronoun: profile.pronoun || GENDERS[google.gender] || "neutral",                                  // 147
  }                                                                                                   // 148
}                                                                                                     // 149
                                                                                                      // 150
function githubIdentity(user, staticHost) {                                                           // 151
  var github = (user.services && user.services.github) || {};                                         // 152
  var profile = user.profile || {};                                                                   // 153
  var id = identityId("github", github.id);                                                           // 154
                                                                                                      // 155
  return {                                                                                            // 156
    service: "github",                                                                                // 157
    id: id,                                                                                           // 158
    name: profile.name || github.username || "Name Unknown",                                          // 159
    email: profile.email || github.email,                                                             // 160
    handle: profile.handle || filterHandle(github.username) ||                                        // 161
            filterHandle(profile.name) || "unknown",                                                  // 162
    picture: staticAssetUrl(profile.picture, staticHost) || makeIdenticon(id),                        // 163
    pronoun: profile.pronoun,                                                                         // 164
  }                                                                                                   // 165
}                                                                                                     // 166
                                                                                                      // 167
function emailIdentity(user, staticHost) {                                                            // 168
  var email = (user.services && user.services.emailToken) || {};                                      // 169
  var profile = user.profile || {};                                                                   // 170
  var id = identityId("email", email.email);                                                          // 171
                                                                                                      // 172
  return {                                                                                            // 173
    service: "email",                                                                                 // 174
    id: id,                                                                                           // 175
    name: profile.name || email.email.split("@")[0] || "Name Unknown",                                // 176
    email: profile.email || email.email,                                                              // 177
    handle: profile.handle || emailToHandle(email.email) ||                                           // 178
            filterHandle(profile.name) || "unknown",                                                  // 179
    picture: staticAssetUrl(profile.picture, staticHost) || makeIdenticon(id),                        // 180
    pronoun: profile.pronoun,                                                                         // 181
  }                                                                                                   // 182
}                                                                                                     // 183
                                                                                                      // 184
function devIdentity(user, staticHost) {                                                              // 185
  var email = (user.services && user.services.emailToken) || {};                                      // 186
  var profile = user.profile || {};                                                                   // 187
  var name = user.devName.split(" ")[0].toLowerCase();                                                // 188
  var id = identityId("dev", user.devName);                                                           // 189
                                                                                                      // 190
  return {                                                                                            // 191
    service: "dev",                                                                                   // 192
    id: id,                                                                                           // 193
    name: profile.name || user.devName,                                                               // 194
    handle: profile.handle || filterHandle(name),                                                     // 195
    picture: staticAssetUrl(profile.picture, staticHost) || makeIdenticon(id),                        // 196
    pronoun: profile.pronoun ||                                                                       // 197
             (_.contains(["alice", "carol", "eve"], name) ? "female" :                                // 198
              _.contains(["bob", "dave"], name) ? "male" : "neutral"),                                // 199
  }                                                                                                   // 200
}                                                                                                     // 201
                                                                                                      // 202
function demoIdentity(user, staticHost) {                                                             // 203
  var email = (user.services && user.services.emailToken) || {};                                      // 204
  var profile = user.profile || {};                                                                   // 205
  var id = identityId("demo", user._id);                                                              // 206
                                                                                                      // 207
  return {                                                                                            // 208
    service: "demo",                                                                                  // 209
    id: id,                                                                                           // 210
    name: profile.name || "Demo User",                                                                // 211
    handle: profile.handle || "demo",                                                                 // 212
    picture: staticAssetUrl(profile.picture, staticHost) || makeIdenticon(id),                        // 213
    pronoun: profile.pronoun || "neutral",                                                            // 214
  }                                                                                                   // 215
}                                                                                                     // 216
                                                                                                      // 217
SandstormDb.getVerifiedEmails = function (user) {                                                     // 218
  var result = [];                                                                                    // 219
                                                                                                      // 220
  var services = user.services || {};                                                                 // 221
  var google = services.google || {};                                                                 // 222
  if (google.email && google.verified_email) result.push(google.email);                               // 223
  if (services.emailToken) result.push(services.emailToken.email);                                    // 224
                                                                                                      // 225
  // TODO(soon): Verification of email addresses -- perhaps through asking the user to log in as      // 226
  //   the given identity?                                                                            // 227
                                                                                                      // 228
  return result;                                                                                      // 229
}                                                                                                     // 230
                                                                                                      // 231
SandstormDb.getUserIdentities = function (user) {                                                     // 232
  // Given a user object, return all of the user's identities.                                        // 233
  //                                                                                                  // 234
  // On the client, must be subscribed "accountIdentities" for the user.                              // 235
                                                                                                      // 236
  var staticHost = httpProtocol + "//" + makeWildcardHost("static");                                  // 237
                                                                                                      // 238
  var result = [];                                                                                    // 239
  if (user && user.services) {                                                                        // 240
    if ("google" in user.services) {                                                                  // 241
      result.push(googleIdentity(user, staticHost));                                                  // 242
    }                                                                                                 // 243
    if ("github" in user.services) {                                                                  // 244
      result.push(githubIdentity(user, staticHost));                                                  // 245
    }                                                                                                 // 246
    if ("emailToken" in user.services) {                                                              // 247
      result.push(emailIdentity(user, staticHost));                                                   // 248
    }                                                                                                 // 249
    if ("devName" in user) {                                                                          // 250
      result.push(devIdentity(user, staticHost));                                                     // 251
    }                                                                                                 // 252
    if ("expires" in user) {                                                                          // 253
      result.push(demoIdentity(user, staticHost));                                                    // 254
    }                                                                                                 // 255
  }                                                                                                   // 256
  return result;                                                                                      // 257
}                                                                                                     // 258
                                                                                                      // 259
////////////////////////////////////////////////////////////////////////////////////////////////////////

}).call(this);






(function () {

////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                    //
// packages/sandstorm-db/user.js                                                                      //
//                                                                                                    //
////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                      //
// Sandstorm - Personal Cloud Sandbox                                                                 // 1
// Copyright (c) 2015 Sandstorm Development Group, Inc. and contributors                              // 2
// All rights reserved.                                                                               // 3
//                                                                                                    // 4
// Licensed under the Apache License, Version 2.0 (the "License");                                    // 5
// you may not use this file except in compliance with the License.                                   // 6
// You may obtain a copy of the License at                                                            // 7
//                                                                                                    // 8
//   http://www.apache.org/licenses/LICENSE-2.0                                                       // 9
//                                                                                                    // 10
// Unless required by applicable law or agreed to in writing, software                                // 11
// distributed under the License is distributed on an "AS IS" BASIS,                                  // 12
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                           // 13
// See the License for the specific language governing permissions and                                // 14
// limitations under the License.                                                                     // 15
                                                                                                      // 16
var Future = Npm.require("fibers/future");                                                            // 17
                                                                                                      // 18
userPictureUrl = function (user) {                                                                    // 19
  if (user.services && !(user.profile && user.profile.picture)) {                                     // 20
    // Try to determine user's avatar URL from login service.                                         // 21
                                                                                                      // 22
    var google = user.services.google;                                                                // 23
    if (google && google.picture) {                                                                   // 24
      return google.picture;                                                                          // 25
    }                                                                                                 // 26
                                                                                                      // 27
    var github = user.services.github;                                                                // 28
    if (github && github.id) {                                                                        // 29
      return "https://avatars.githubusercontent.com/u/" + github.id;                                  // 30
    }                                                                                                 // 31
                                                                                                      // 32
    // Note that we do NOT support Gravatar for email addresses because pinging Gravatar would be     // 33
    // a data leak, revealing that the user has logged into this Sandstorm server. Google and         // 34
    // Github are different because they are actually the identity providers, so they already know    // 35
    // the user logged in.                                                                            // 36
  }                                                                                                   // 37
}                                                                                                     // 38
                                                                                                      // 39
fetchPicture = function (url) {                                                                       // 40
  try {                                                                                               // 41
    var result = HTTP.get(url, {                                                                      // 42
      npmRequestOptions: { encoding: null },                                                          // 43
      timeout: 5000                                                                                   // 44
    });                                                                                               // 45
                                                                                                      // 46
    var metadata = {};                                                                                // 47
                                                                                                      // 48
    metadata.mimeType = result.headers["content-type"];                                               // 49
    if (metadata.mimeType.lastIndexOf("image/png", 0) === -1 &&                                       // 50
        metadata.mimeType.lastIndexOf("image/jpeg", 0) === -1) {                                      // 51
      throw new Error("unexpected Content-Type:", metadata.mimeType);                                 // 52
    }                                                                                                 // 53
    var enc = result.headers["content-encoding"];                                                     // 54
    if (enc && enc !== "identity") {                                                                  // 55
      metadata.encoding = enc;                                                                        // 56
    }                                                                                                 // 57
                                                                                                      // 58
    return addStaticAsset(metadata, result.content);                                                  // 59
  } catch (err) {                                                                                     // 60
    console.error("failed to fetch user profile picture:", url, err.stack);                           // 61
  }                                                                                                   // 62
}                                                                                                     // 63
                                                                                                      // 64
var ValidHandle = Match.Where(function (handle) {                                                     // 65
  check(handle, String);                                                                              // 66
  return !!handle.match(/^[a-z_][a-z0-9_]*$/);                                                        // 67
});                                                                                                   // 68
                                                                                                      // 69
Accounts.onCreateUser(function (options, user) {                                                      // 70
  // The first non-dev user to sign in should be automatically upgraded to admin.                     // 71
  // Dev users are identified by having the devName field.                                            // 72
  if (Meteor.users.find({devName: {$exists: 0}}).count() === 0 && !user.devName) {                    // 73
    user.isAdmin = true;                                                                              // 74
    user.signupKey = "admin";                                                                         // 75
  }                                                                                                   // 76
                                                                                                      // 77
  // Check profile.                                                                                   // 78
  if (options.profile) {                                                                              // 79
    // TODO(cleanup): This check also appears in accounts-ui-methods.js.                              // 80
    check(options.profile, Match.ObjectIncluding({                                                    // 81
      name: Match.OneOf(null, Match.Optional(String)),                                                // 82
      handle: Match.Optional(ValidHandle),                                                            // 83
      pronoun: Match.Optional(Match.OneOf("male", "female", "neutral", "robot")),                     // 84
    }));                                                                                              // 85
                                                                                                      // 86
    user.profile = options.profile;                                                                   // 87
  } else {                                                                                            // 88
    user.profile = {};                                                                                // 89
  }                                                                                                   // 90
                                                                                                      // 91
  // Try downloading avatar.                                                                          // 92
  var url = userPictureUrl(user);                                                                     // 93
  if (url) {                                                                                          // 94
    var assetId = fetchPicture(url);                                                                  // 95
    if (assetId) {                                                                                    // 96
      user.profile.picture = assetId;                                                                 // 97
    }                                                                                                 // 98
  }                                                                                                   // 99
                                                                                                      // 100
  return user;                                                                                        // 101
});                                                                                                   // 102
                                                                                                      // 103
////////////////////////////////////////////////////////////////////////////////////////////////////////

}).call(this);






(function () {

////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                    //
// packages/sandstorm-db/migrations.js                                                                //
//                                                                                                    //
////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                      //
// Minimal tooling for doing run-at-least-once, ordered migrations.                                   // 1
//                                                                                                    // 2
// Because migrations can experience partial failure and likely have                                  // 3
// side-effects, we should be careful to make sure all migrations are                                 // 4
// idempotent and safe to accidentally run multiple times.                                            // 5
                                                                                                      // 6
var Future = Npm.require("fibers/future");                                                            // 7
                                                                                                      // 8
var updateLoginStyleToRedirect = function() {                                                         // 9
  var configurations = Package["service-configuration"].ServiceConfiguration.configurations;          // 10
  ["google", "github"].forEach(function(serviceName) {                                                // 11
    var config = configurations.findOne({service: serviceName});                                      // 12
    if (config && config.loginStyle !== "redirect") {                                                 // 13
      configurations.update({service: serviceName}, {$set: {loginStyle: "redirect"}});                // 14
    }                                                                                                 // 15
  });                                                                                                 // 16
};                                                                                                    // 17
                                                                                                      // 18
var enableLegacyOAuthProvidersIfNotInSettings = function() {                                          // 19
  // In the before time, Google and Github login were enabled by default.                             // 20
  //                                                                                                  // 21
  // This actually didn't make much sense, required the first user to configure                       // 22
  // OAuth, and had some trust-the-first-user properties that weren't totally                         // 23
  // secure.                                                                                          // 24
  //                                                                                                  // 25
  // Now, we have admin-token, so we wish to disable all logins by default                            // 26
  // (since they need to be configured anyway) but since users may have never                         // 27
  // explicitly told Sandstorm that Google or Github login should be enabled,                         // 28
  // we can't just use the value in the Settings collection, since it might                           // 29
  // never have been set.                                                                             // 30
  //                                                                                                  // 31
  // Thus, if the service is configured but absent from Settings, we should                           // 32
  // explicitly enable it in Settings, and then the rest of the logic can just                        // 33
  // depend on what value is in Settings and default to false without breaking                        // 34
  // user installations.                                                                              // 35
  var configurations = Package["service-configuration"].ServiceConfiguration.configurations;          // 36
  ["google", "github"].forEach(function(serviceName) {                                                // 37
    var config = configurations.findOne({service: serviceName});                                      // 38
    var serviceConfig = Settings.findOne({_id: serviceName});                                         // 39
    if (config && !serviceConfig) {                                                                   // 40
      // Only explicitly enable the login service if:                                                 // 41
      // 1) the service is already configured                                                         // 42
      // 2) there is no sandstorm configuration already present (the user was                         // 43
      //    using the previous default behavior).                                                     // 44
      Settings.insert({_id: serviceName, value: true});                                               // 45
    }                                                                                                 // 46
  });                                                                                                 // 47
};                                                                                                    // 48
                                                                                                      // 49
var denormalizeInviteInfo = function() {                                                              // 50
  // When a user is invited via a signup token, the `signupKey` field of their user table entry       // 51
  // has always been populated to indicate the key they used. This points into the SignupKeys table   // 52
  // which has more information about the key, namely a freeform note entered by the admin when       // 53
  // they created the key. In the case that the email invite form was used, the note has the form     // 54
  // "E-mail invite to <address>".                                                                    // 55
  //                                                                                                  // 56
  // Later, we decided it was useful to indicate in the users table visible to the admin              // 57
  // information about the invite terms. Namely, for email invites we want to show the address        // 58
  // and for others we want to show the note. To make this efficient, fields `signupNote` and         // 59
  // `signupEmail` were added to the users table. We can backfill these values by denormalizing       // 60
  // from the SignupKeys table.                                                                       // 61
                                                                                                      // 62
  Meteor.users.find().forEach(function (user) {                                                       // 63
    if (user.signupKey && (typeof user.signupKey) === "string" && user.signupKey !== "admin") {       // 64
      var signupInfo = SignupKeys.findOne(user.signupKey);                                            // 65
      if (signupInfo && signupInfo.note) {                                                            // 66
        var newFields = { signupNote: signupInfo.note };                                              // 67
                                                                                                      // 68
        var prefix = "E-mail invite to ";                                                             // 69
        if (signupInfo.note.lastIndexOf(prefix) === 0) {                                              // 70
          newFields.signupEmail = signupInfo.note.slice(prefix.length);                               // 71
        }                                                                                             // 72
                                                                                                      // 73
        Meteor.users.update(user._id, {$set: newFields});                                             // 74
      }                                                                                               // 75
    }                                                                                                 // 76
  });                                                                                                 // 77
}                                                                                                     // 78
                                                                                                      // 79
function mergeRoleAssignmentsIntoApiTokens() {                                                        // 80
  RoleAssignments.find().forEach(function(roleAssignment) {                                           // 81
    ApiTokens.insert({                                                                                // 82
      grainId: roleAssignment.grainId,                                                                // 83
      userId: roleAssignment.sharer,                                                                  // 84
      roleAssignment: roleAssignment.roleAssignment,                                                  // 85
      petname: roleAssignment.petname,                                                                // 86
      created: roleAssignment.created,                                                                // 87
      owner: {user: {userId: roleAssignment.recipient,                                                // 88
                     title: roleAssignment.title}},                                                   // 89
    });                                                                                               // 90
  });                                                                                                 // 91
}                                                                                                     // 92
                                                                                                      // 93
function fixOasisStorageUsageStats() {                                                                // 94
  // Before storage quota enforcement was first implemented, there were some bugs with the way the    // 95
  // Oasis backend tracked storage usage, which were not initially noticed since the resulting        // 96
  // number wasn't acutally used for anything.                                                        // 97
  //                                                                                                  // 98
  // With the underlying bugs fixed, simply starting up each grain (supervisor only, not the app)     // 99
  // and then immediately shutting it down should cause the sizes to be recalculated correctly.       // 100
  //                                                                                                  // 101
  // This problem only applies to Oasis.                                                              // 102
                                                                                                      // 103
  if (Meteor.settings.public.quotaEnabled) {                                                          // 104
    var dummyCommand = {                                                                              // 105
      argv: ["foo", "bar"],                                                                           // 106
      environ: []                                                                                     // 107
    };                                                                                                // 108
                                                                                                      // 109
    Grains.find().forEach(function (grain) {                                                          // 110
      try {                                                                                           // 111
        console.log("refreshing grain:", grain._id);                                                  // 112
                                                                                                      // 113
        // We take a highly synchronous approach here so that the logging is understandable.          // 114
        var supervisor = waitPromise(sandstormBackend.startGrain(                                     // 115
            grain.userId, grain._id, grain.packageId, dummyCommand, false, false)).supervisor;        // 116
        try {                                                                                         // 117
          waitPromise(supervisor.shutdown());                                                         // 118
          throw new Error("shutdown() didn't disconnect!");                                           // 119
        } catch (err) {                                                                               // 120
          if (err.kjType !== "disconnected") throw err;                                               // 121
        }                                                                                             // 122
      } catch (err) {                                                                                 // 123
        console.log(err.stack);                                                                       // 124
      }                                                                                               // 125
    });                                                                                               // 126
  }                                                                                                   // 127
}                                                                                                     // 128
                                                                                                      // 129
function fetchProfilePictures() {                                                                     // 130
  Meteor.users.find({}).forEach(function (user) {                                                     // 131
    var url = userPictureUrl(user);                                                                   // 132
    if (url) {                                                                                        // 133
      console.log("Fetching user picture:", url);                                                     // 134
      var assetId = fetchPicture(url);                                                                // 135
      if (assetId) {                                                                                  // 136
        Meteor.users.update(user._id, {$set: {"profile.picture": assetId}});                          // 137
      }                                                                                               // 138
    }                                                                                                 // 139
  });                                                                                                 // 140
}                                                                                                     // 141
                                                                                                      // 142
function assignPlans() {                                                                              // 143
  if (Meteor.settings.public.quotaEnabled && SandstormDb.paymentsMigrationHook) {                     // 144
    SandstormDb.paymentsMigrationHook(SignupKeys, Plans.find().fetch());                              // 145
  }                                                                                                   // 146
}                                                                                                     // 147
                                                                                                      // 148
function removeKeyrings() {                                                                           // 149
  // These blobs full of public keys were not intended to find their way into mongo and while         // 150
  // harmless they slow things down because they're huge. Remove them.                                // 151
  Packages.update({"manifest.metadata.pgpKeyring": {$exists: true}},                                  // 152
      {$unset: {"manifest.metadata.pgpKeyring": ""}},                                                 // 153
      {multi: true});                                                                                 // 154
}                                                                                                     // 155
                                                                                                      // 156
// This must come after all the functions named within are defined.                                   // 157
// Only append to this list!  Do not modify or remove list entries;                                   // 158
// doing so is likely change the meaning and semantics of user databases.                             // 159
var MIGRATIONS = [                                                                                    // 160
  updateLoginStyleToRedirect,                                                                         // 161
  enableLegacyOAuthProvidersIfNotInSettings,                                                          // 162
  denormalizeInviteInfo,                                                                              // 163
  mergeRoleAssignmentsIntoApiTokens,                                                                  // 164
  fixOasisStorageUsageStats,                                                                          // 165
  fetchProfilePictures,                                                                               // 166
  assignPlans,                                                                                        // 167
  removeKeyrings,                                                                                     // 168
];                                                                                                    // 169
                                                                                                      // 170
function migrateToLatest() {                                                                          // 171
  if (Meteor.settings.replicaNumber) {                                                                // 172
    // This is a replica. Wait for the first replica to perform migrations.                           // 173
                                                                                                      // 174
    console.log("Waiting for migrations on replica zero...");                                         // 175
                                                                                                      // 176
    var done = new Future();                                                                          // 177
    var change = function (doc) {                                                                     // 178
      console.log("Migrations applied elsewhere: " + doc.value + "/" + MIGRATIONS.length);            // 179
      if (doc.value >= MIGRATIONS.length) done.return();                                              // 180
    }                                                                                                 // 181
    var observer = Migrations.find({_id: "migrations_applied"}).observe({                             // 182
      added: change,                                                                                  // 183
      changed: change                                                                                 // 184
    });                                                                                               // 185
                                                                                                      // 186
    done.wait();                                                                                      // 187
    observer.stop();                                                                                  // 188
    console.log("Migrations have completed on replica zero.");                                        // 189
                                                                                                      // 190
  } else {                                                                                            // 191
    var applied = Migrations.findOne({_id: "migrations_applied"});                                    // 192
    var start;                                                                                        // 193
    if (!applied) {                                                                                   // 194
      // Migrations table is not yet seeded with a value.  This means it has                          // 195
      // applied 0 migrations.  Persist this.                                                         // 196
      Migrations.insert({_id: "migrations_applied", value: 0});                                       // 197
      start = 0;                                                                                      // 198
    } else {                                                                                          // 199
      start = applied.value;                                                                          // 200
    }                                                                                                 // 201
    console.log("Migrations applied: " + start + "/" + MIGRATIONS.length);                            // 202
                                                                                                      // 203
    for (var i = start ; i < MIGRATIONS.length ; i++) {                                               // 204
      // Apply migration i, then record that migration i was successfully run.                        // 205
      console.log("Applying migration " + (i+1));                                                     // 206
      MIGRATIONS[i]();                                                                                // 207
      Migrations.update({_id: "migrations_applied"}, {$set: {value: i+1}});                           // 208
      console.log("Applied migration " + (i+1));                                                      // 209
    }                                                                                                 // 210
  }                                                                                                   // 211
}                                                                                                     // 212
                                                                                                      // 213
// Apply all migrations on startup.                                                                   // 214
Meteor.startup(migrateToLatest);                                                                      // 215
                                                                                                      // 216
////////////////////////////////////////////////////////////////////////////////////////////////////////

}).call(this);


/* Exports */
if (typeof Package === 'undefined') Package = {};
Package['sandstorm-db'] = {
  SandstormDb: SandstormDb
};

})();

//# sourceMappingURL=sandstorm-db.js.map
